#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom errors.
"""


class CPIObjectDoesNotExist(Exception):
    """
    Error raised when a CPI object is requested that doesn't exist.
    """
    pass


class StaleDataWarning(Warning):
    """
    The warning to raise when the local data are out of date.
    """
    def __str__(self):
        return "CPI data is out of date. To accurately inflate to today's dollars, you must run `cpi.update()`."



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Command-line interface.
"""
import cpi
import click
from dateutil.parser import parse as dateparse


@click.command()
@click.argument(
    "value",
    nargs=1,
    required=True,
)
@click.argument(
    "year_or_month",
    nargs=1,
    required=True,
)
@click.option(
    '--to',
    nargs=1,
    default=None,
    help="The year or month to adjust the value to."
)
@click.option(
    '--series',
    type=click.STRING,
    nargs=1,
    default=cpi.DEFAULT_SERIES_ID,
    help='The CPI data series used for the conversion. The default is the CPI-U.'
)
def inflate(value, year_or_month, to=None, series=cpi.DEFAULT_SERIES):
    """
    Returns a dollar value adjusted for inflation.
    """
    # Sanitize the value
    try:
        value = float(value)
    except ValueError:
        click.ClickException("Dollar value must be an integer or float.")

    # Sanitize the `from` date.
    try:
        year_or_month = _parse_date(year_or_month)
    except ValueError:
        click.ClickException("Source date must be a year as an integer or a month as a parseable date string.")

    # Sanitize the `to` date.
    if to:
        try:
            to = _parse_date(to)
        except ValueError:
            click.ClickException("Source date must be a year as an integer or a month as a parseable date string.")

    # Run the command
    result = cpi.inflate(value, year_or_month, to=to, series=series)

    # Print out the result to the terminal
    click.echo(result)


def _parse_date(value):
    """
    Parse a date submitted to the CLIself.

    Returns and integer if its a year. Returns a date object if its a month.
    """
    try:
        return int(value)
    except ValueError:
        return dateparse(value).date()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Python objects for modeling Consumer Price Index (CPI) data structures.
"""
import logging
import collections
from datetime import date
from .errors import CPIObjectDoesNotExist
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MappingList(list):
    """
    A custom list that allows for lookups by attribute.
    """
    def get_by_id(self, value):
        try:
            return list(filter(lambda obj: obj.id == value, self))[0]
        except IndexError:
            raise CPIObjectDoesNotExist("Object with id {} could not be found".format(value))

    def get_by_name(self, value):
        try:
            return list(filter(lambda obj: obj.name == value, self))[0]
        except IndexError:
            raise CPIObjectDoesNotExist("Object with id {} could not be found".format(value))


class SeriesList(list):
    """
    A custom list of indexes in a series.
    """
    SURVEYS = {
        'All urban consumers': 'CU',
        'Urban wage earners and clerical workers': 'CW'
    }
    SEASONALITIES = {
        True: 'S',
        False: 'U'
    }

    def __init__(self, periodicities, areas, items):
        self.periodicities = periodicities
        self.areas = areas
        self.items = items
        self._dict = {}

    def append(self, item):
        """
        Override to default append method that allows validation and dictionary-style lookups
        """
        # Valid item type
        if not isinstance(item, Series):
            raise TypeError("Only Series objects can be added to this list.")

        # Add to dictionary lookup
        self._dict[item.id] = item

        # Append to list
        super(SeriesList, self).append(item)

    def get_by_id(self, value):
        logger.debug("Retrieving series with id {}".format(value))
        try:
            return self._dict[value]
        except KeyError:
            raise CPIObjectDoesNotExist("Object with id {} could not be found".format(value))

    def get(self, survey, seasonally_adjusted, periodicity, area, items):
        # Get all the codes for these humanized input.
        try:
            survey_code = self.SURVEYS[survey]
        except KeyError:
            raise CPIObjectDoesNotExist("Survey with the name {} does not exist".format(survey))

        try:
            seasonality_code = self.SEASONALITIES[seasonally_adjusted]
        except KeyError:
            raise CPIObjectDoesNotExist("Seasonality {} does not exist".format(seasonally_adjusted))

        # Generate the series id
        series_id = "{}{}{}{}{}".format(
            survey_code,
            seasonality_code,
            self.periodicities.get_by_name(periodicity).code,
            self.areas.get_by_name(area).code,
            self.items.get_by_name(items).code
        )

        # Pull the series
        return self.get_by_id(series_id)


class BaseObject(object):
    """
    An abstract base class for all the models.
    """
    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.__str__())


class Area(BaseObject):
    """
    A geographical area where prices are gathered monthly.
    """
    def __init__(self, code, name):
        self.id = code
        self.code = code
        self.name = name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id


class Item(BaseObject):
    """
    A consumer item that has its price tracked.
    """
    def __init__(self, code, name):
        self.id = code
        self.code = code
        self.name = name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id


class Period(BaseObject):
    """
    A time period tracked by the CPI.
    """
    def __init__(self, code, abbreviation, name):
        self.id = code
        self.code = code
        self.abbreviation = abbreviation
        self.name = name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id

    @property
    def month(self):
        """
        Returns the month integer for the period.
        """
        if self.id in ["M13", "S01", "S03"]:
            return 1
        elif self.id == "S02":
            return 7
        else:
            return int(self.id.replace("M", ""))

    @property
    def type(self):
        """
        Returns a string classifying the period.
        """
        if self.id in ["M13", "S03"]:
            return "annual"
        elif self.id in ["S01", "S02"]:
            return "semiannual"
        else:
            return "monthly"


class Periodicity(BaseObject):
    """
    A time interval tracked by the CPI.
    """
    def __init__(self, code, name):
        self.id = code
        self.code = code
        self.name = name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id


class Series(BaseObject):
    """
    A set of CPI data observed over an extended period of time over consistent time intervals ranging from
    a specific consumer item in a specific geographical area whose price is gathered monthly to a category
    of worker in a specific industry whose employment rate is being recorded monthly, etc.

    Yes, that's the offical government definition. I'm not kidding.
    """
    def __init__(
        self,
        id,
        title,
        survey,
        seasonally_adjusted,
        periodicity,
        area,
        items
    ):
        self.id = id
        self.title = title
        self.survey = survey
        self.seasonally_adjusted = seasonally_adjusted
        self.periodicity = periodicity
        self.area = area
        self.items = items
        self._indexes = {
            'annual': collections.OrderedDict(),
            'monthly': collections.OrderedDict(),
            'semiannual': collections.OrderedDict(),
        }

    def __str__(self):
        return "{}: {}".format(self.id, self.title)

    def __eq__(self, other):
        return self.id == other.id

    @property
    def indexes(self):
        flat = []
        for l in self._indexes.values():
            flat.extend(l)
        return flat

    @property
    def latest_month(self):
        if not self._indexes['monthly']:
            return None
        return max([i.date for i in self._indexes['monthly'].values()])

    @property
    def latest_year(self):
        if not self._indexes['annual']:
            return None
        return max([i.year for i in self._indexes['annual'].values()])

    def get_index_by_date(self, date, period_type='annual'):
        try:
            return self._indexes[period_type][date]
        except KeyError:
            raise CPIObjectDoesNotExist("Index of {} type for {} does not exist".format(period_type, date))


class Index(BaseObject):
    """
    A Consumer Price Index value generated by the Bureau of Labor Statistics.
    """
    def __init__(self, series, year, period, value):
        self.series = series
        self.year = year
        self.period = period
        self.value = value

    def __str__(self):
        return "{} ({}): {}".format(self.date, self.period, self.value)

    def __eq__(self, other):
        return (
            self.value == other.value and
            self.series == other.series and
            self.year == other.year and
            self.period == other.period
        )

    @property
    def date(self):
        """
        Accepts a row from the raw BLS data. Returns a Python date object based on its period.
        """
        return date(self.year, self.period.month, 1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Default values.
"""
DEFAULT_SERIES_ID = "CUUR0000SA0"
DEFAULTS_SERIES_ATTRS = {
    'survey': 'All urban consumers',
    'seasonally_adjusted': False,
    'periodicity': 'Monthly',
    'area': 'U.S. city average',
    'items': 'All items'
}



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Download the latest annual Consumer Price Index (CPI) dataset.
"""
import os
import csv
import sqlite3
import logging
import requests
import pandas as pd
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Downloader(object):
    THIS_DIR = os.path.dirname(__file__)
    FILE_LIST = [
        "cu.area",
        "cu.item",
        "cu.period",
        "cu.periodicity",
        "cu.series",
        "cu.data.0.Current",
        "cu.data.1.AllItems",
        "cu.data.2.Summaries",
        "cu.data.3.AsizeNorthEast",
        "cu.data.4.AsizeNorthCentral",
        "cu.data.5.AsizeSouth",
        "cu.data.6.AsizeWest",
        "cu.data.7.OtherNorthEast",
        "cu.data.8.OtherNorthCentral",
        "cu.data.9.OtherSouth",
        "cu.data.10.OtherWest",
        "cu.data.11.USFoodBeverage",
        "cu.data.12.USHousing",
        "cu.data.13.USApparel",
        "cu.data.14.USTransportation",
        "cu.data.15.USMedical",
        "cu.data.16.USRecreation",
        "cu.data.17.USEducationAndCommunication",
        "cu.data.18.USOtherGoodsAndServices",
        "cu.data.19.PopulationSize",
        "cu.data.20.USCommoditiesServicesSpecial"
    ]

    def update(self):
        """
        Update the Consumer Price Index dataset that powers this library.
        """
        logger.debug("Downloading {} files from the BLS".format(len(self.FILE_LIST)))
        [self.get_tsv(file) for file in self.FILE_LIST]
        logger.debug("Loading data into SQLite database")
        [self.insert_tsv(file) for file in self.FILE_LIST]

    def insert_tsv(self, file):
        # Connect to db
        db_path = os.path.join(self.THIS_DIR, 'cpi.db')
        conn = sqlite3.connect(db_path)

        # Read file
        logger.debug(" - {}".format(file))
        csv_path = os.path.join(self.THIS_DIR, '{}.csv'.format(file))
        csv_reader = list(csv.DictReader(open(csv_path, 'r')))

        # Convert it to a DataFrame
        df = pd.DataFrame(csv_reader)
        df.drop([None], axis=1, inplace=True, errors="ignore")

        # Write file to db
        df.to_sql(file, conn, if_exists="replace", index=False)

        # Close connection
        conn.close()

    def get_tsv(self, file):
        """
        Download TSV file from the BLS.
        """
        # Download it
        url = "https://download.bls.gov/pub/time.series/cu/{}".format(file)
        logger.debug(" - {}".format(url))
        tsv_path = os.path.join(self.THIS_DIR, '{}.tsv'.format(file))
        response = requests.get(url)
        with open(tsv_path, 'w') as f:
            f.write(response.text)

        # Convert it to csv
        reader = csv.reader(open(tsv_path, 'r'), delimiter="\t")
        csv_path = os.path.join(self.THIS_DIR, '{}.csv'.format(file))
        writer = csv.writer(open(csv_path, 'w'))
        for row in reader:
            writer.writerow([cell.strip() for cell in row])


if __name__ == '__main__':
    Downloader().update()



import os
from setuptools import setup
from distutils.core import Command


setup(
    name='cpi',
    version='0.1.1',
    description="Quickly adjust U.S. dollars for inflation using the Consumer Price Index (CPI)",
    author='Ben Welsh',
    author_email='ben.welsh@gmail.com',
    url='http://www.github.com/datadesk/cpi',
    license="MIT",
    packages=("cpi",),
    include_package_data=True,
    zip_safe=False,  # because we're including static files
    install_requires=(
        "requests>=2.19.1",
        "click>=6.7",
        "python-dateutil>=2.7.3",
    ),
    entry_points="""
        [console_scripts]
        inflate=cpi.cli:inflate
    """,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
    ],
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Parse and prepare the Consumer Price Index (CPI) dataset.
"""
import os
import sqlite3
import logging
from .models import MappingList, SeriesList
from .models import Area, Item, Period, Periodicity, Index, Series
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class BaseParser(object):
    THIS_DIR = os.path.dirname(__file__)

    def get_file(self, file):
        """
        Returns the CPI data as a csv.DictReader object.
        """
        # Connect to database
        db_path = os.path.join(self.THIS_DIR, "cpi.db")
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()

        # Query this file
        query = cursor.execute('SELECT * FROM "{}"'.format(file))
        columns = [d[0] for d in query.description]
        result_list = [dict(zip(columns, r)) for r in query.fetchall()]

        # Close database connection
        cursor.close()
        cursor.connection.close()

        # Return data
        return result_list


class ParseArea(BaseParser):
    """
    Parses the raw list of CPI areas.
    """
    def parse(self):
        """
        Returns a list Area objects.
        """
        logger.debug("Parsing area file")
        object_list = MappingList()
        for row in self.get_file('cu.area'):
            obj = Area(row['area_code'], row['area_name'])
            object_list.append(obj)
        return object_list


class ParseItem(BaseParser):
    """
    Parses the raw list of CPI items.
    """
    def parse(self):
        """
        Returns a list Area objects.
        """
        logger.debug("Parsing item file")
        object_list = MappingList()
        for row in self.get_file('cu.item'):
            obj = Item(row['item_code'], row['item_name'])
            object_list.append(obj)
        return object_list


class ParsePeriod(BaseParser):
    """
    Parses the raw list of CPI periods.
    """
    def parse(self):
        """
        Returns a list Area objects.
        """
        logger.debug("Parsing period file")
        object_list = MappingList()
        for row in self.get_file('cu.period'):
            obj = Period(row['period'], row['period_abbr'], row['period_name'])
            object_list.append(obj)
        return object_list


class ParsePeriodicity(BaseParser):
    """
    Parses the raw list of CPI periodicities.
    """
    def parse(self):
        """
        Returns a list Periodicity objects.
        """
        logger.debug("Parsing periodicity file")
        object_list = MappingList()
        for row in self.get_file('cu.periodicity'):
            obj = Periodicity(row['periodicity_code'], row['periodicity_name'])
            object_list.append(obj)
        return object_list


class ParseSeries(BaseParser):
    """
    Parses the raw list of CPI series from the BLS.
    """
    SURVEYS = {
        'CU': 'All urban consumers',
        'CW': 'Urban wage earners and clerical workers'
    }
    FILE_LIST = [
        "cu.data.0.Current",
        "cu.data.1.AllItems",
        "cu.data.2.Summaries",
        "cu.data.3.AsizeNorthEast",
        "cu.data.4.AsizeNorthCentral",
        "cu.data.5.AsizeSouth",
        "cu.data.6.AsizeWest",
        "cu.data.7.OtherNorthEast",
        "cu.data.8.OtherNorthCentral",
        "cu.data.9.OtherSouth",
        "cu.data.10.OtherWest",
        "cu.data.11.USFoodBeverage",
        "cu.data.12.USHousing",
        "cu.data.13.USApparel",
        "cu.data.14.USTransportation",
        "cu.data.15.USMedical",
        "cu.data.16.USRecreation",
        "cu.data.17.USEducationAndCommunication",
        "cu.data.18.USOtherGoodsAndServices",
        "cu.data.19.PopulationSize",
        "cu.data.20.USCommoditiesServicesSpecial"
    ]

    def __init__(self, periods=None, periodicities=None, areas=None, items=None):
        self.periods = periods or ParsePeriod().parse()
        self.periodicities = periodicities or ParsePeriodicity().parse()
        self.areas = areas or ParseArea().parse()
        self.items = items or ParseItem().parse()

    def parse_id(self, id):
        return dict(
            survey_code=id[:2],
            seasonal_code=id[2:3],
            periodicity_code=id[3:4],
            area_code=id[4:8],
            item_code=id[8:]
        )

    def parse(self):
        self.series_list = self.parse_series()
        self.parse_indexes()
        return self.series_list

    def parse_series(self):
        """
        Returns a list Series objects.
        """
        logger.debug("Parsing series file")
        object_list = SeriesList(
            periodicities=self.periodicities,
            areas=self.areas,
            items=self.items
        )
        for row in self.get_file('cu.series'):
            parsed_id = self.parse_id(row['series_id'])
            obj = Series(
                row['series_id'],
                row['series_title'],
                self.SURVEYS[parsed_id['survey_code']],
                row['seasonal'] == 'S',
                self.periodicities.get_by_id(row['periodicity_code']),
                self.areas.get_by_id(row['area_code']),
                self.items.get_by_id(row['item_code'])
            )
            object_list.append(obj)
        return object_list

    def parse_indexes(self):
        logger.debug("Parsing index files")
        # Loop through all the files ...
        for file in self.FILE_LIST:
            # ... and for each file ...
            for row in self.get_file(file):
                # Get the series
                series = self.series_list.get_by_id(row['series_id'])

                # Create an object
                index = Index(
                    series,
                    int(row['year']),
                    self.periods.get_by_id(row['period']),
                    float(row['value'])
                )

                # If the value has already been loaded ...
                if index.date in series._indexes[index.period.type]:
                    # ... verify this value matches what we have ...
                    assert index == series._indexes[index.period.type][index.date]
                else:
                    # ... and if the series doesn't have the index yet, add it.
                    series._indexes[index.period.type][index.date] = index


def parse():
    """
    Parse all of the files and returns a ready-to-use series list.
    """
    logger.info("Parsing data files from the BLS")
    areas = ParseArea().parse()
    items = ParseItem().parse()
    periods = ParsePeriod().parse()
    periodicities = ParsePeriodicity().parse()
    series_list = ParseSeries(
        periods=periods,
        periodicities=periodicities,
        areas=areas,
        items=items
    ).parse()
    return series_list



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Quickly adjust U.S. dollars for inflation using the Consumer Price Index (CPI)
"""
import numbers
import warnings
from datetime import date, datetime

from .parsers import parse
from .download import Downloader
from .errors import StaleDataWarning
from .defaults import DEFAULT_SERIES_ID, DEFAULTS_SERIES_ATTRS

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Parse data for use
SERIES_LIST = parse()

# set the default series to the CPI-U
DEFAULT_SERIES = SERIES_LIST.get_by_id(DEFAULT_SERIES_ID)

# Establish the range of data available
LATEST_MONTH = DEFAULT_SERIES.latest_month
LATEST_YEAR = DEFAULT_SERIES.latest_year

# Figure out how out of date you are
DAYS_SINCE_LATEST_MONTH = (date.today() - LATEST_MONTH).days
DAYS_SINCE_LATEST_YEAR = (date.today() - date(LATEST_YEAR, 1, 1)).days

# If it's more than two and a half years out of date, raise a warning.
if DAYS_SINCE_LATEST_YEAR > (365*2.25) or DAYS_SINCE_LATEST_MONTH > 90:
    warnings.warn(StaleDataWarning())
    logger.warn("CPI data is out of date. To accurately inflate to today's dollars, you must run `cpi.update()`.")


def get(
    year_or_month,
    survey=DEFAULTS_SERIES_ATTRS['survey'],
    seasonally_adjusted=DEFAULTS_SERIES_ATTRS['seasonally_adjusted'],
    periodicity=DEFAULTS_SERIES_ATTRS['periodicity'],
    area=DEFAULTS_SERIES_ATTRS['area'],
    items=DEFAULTS_SERIES_ATTRS['items'],
    series=None
):
    """
    Returns the CPI value for a given year.
    """
    # Prep the lookup value depending on the input type.
    if isinstance(year_or_month, numbers.Integral):
        year_or_month = date(year_or_month, 1, 1)
        period_type = 'annual'
    elif isinstance(year_or_month, date):
        period_type = 'monthly'
        # If it's not set to the first day of the month, we should do that now.
        if year_or_month.day != 1:
            year_or_month = year_or_month.replace(day=1)
    else:
        raise ValueError("Only integers and date objects are accepted.")

    # Pull the series
    if series:
        # If the user has provided an explicit series id, we are going to ignore the humanized options.
        series = SERIES_LIST.get_by_id(series)
    else:
        # Otherwise, we build the series id using the more humanized options
        series = SERIES_LIST.get(survey, seasonally_adjusted, periodicity, area, items)

    # Pull the value from the series by date
    return series.get_index_by_date(year_or_month, period_type=period_type).value


def inflate(
    value,
    year_or_month,
    to=None,
    survey=DEFAULTS_SERIES_ATTRS['survey'],
    seasonally_adjusted=DEFAULTS_SERIES_ATTRS['seasonally_adjusted'],
    periodicity=DEFAULTS_SERIES_ATTRS['periodicity'],
    area=DEFAULTS_SERIES_ATTRS['area'],
    items=DEFAULTS_SERIES_ATTRS['items'],
    series=None
):
    """
    Returns a dollar value adjusted for inflation.

    You must submit the value, followed by the year or month its from.

    Years should be submitted as integers. Months as datetime.date objects.

    By default, the input is adjusted to the most recent year or month available from the CPI.

    If you'd like to adjust to a different year or month, submit it to the optional `to` keyword argument.

    Yearly data can only be updated to other years. Monthly data can only be updated to other months.
    """
    # If the two dates match, just return the value unadjusted
    if year_or_month == to:
        return value

    # Figure out the 'to' date if it has not been provided
    if not to:
        if isinstance(year_or_month, (date, datetime)):
            to = LATEST_MONTH
        else:
            to = LATEST_YEAR
    # Otherwise sanitize it
    else:
        if isinstance(to, numbers.Integral):
            to = int(to)
        elif isinstance(to, datetime):
            # We want dates not datetimes
            to = to.date()

    # Sanitize the year_or_month
    if isinstance(year_or_month, numbers.Integral):
        # We need to make sure that int64, int32 and other int-like objects
        # are the same type for the comparison to come.
        year_or_month = int(year_or_month)
    # If a datetime has been provided, shave it down to a date.
    elif isinstance(year_or_month, datetime):
        year_or_month = year_or_month.date()

    # Make sure the two dates are the same type
    if type(year_or_month) != type(to):
        raise TypeError("Years can only be converted to other years. Months only to other months.")

    # Otherwise, let's do the math.
    # The input value is multiplied by the CPI of the target year,
    # then divided into the CPI from the source year.
    kwargs = {
        'survey': survey,
        'seasonally_adjusted': seasonally_adjusted,
        'periodicity': periodicity,
        'area': area,
        'items': items,
        'series': series
    }
    source_index = get(year_or_month, **kwargs)
    target_index = get(to, **kwargs)
    return (value * target_index) / float(source_index)


def update():
    """
    Updates the core Consumer Price Index dataset at the core of this library.

    Requires an Internet connection.
    """
    Downloader().update()
